# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2017 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import xml.etree.ElementTree as ET
from django.utils.text import slugify

from django.db import models
from django.utils.translation import ugettext_lazy as _

from passerelle.base.models import BaseResource
from passerelle.utils.api import endpoint

def build_dict_from_xml(elem):
    d = {}
    for child in elem.find('.'):
        if child.tag.startswith('{'):
            continue
        attribute_name = slugify(child.tag).replace('-', '_')
        if child.text and child.text.strip():
            d[attribute_name] = child.text.strip()
        else:
            d[attribute_name] = build_dict_from_xml(child)
    return d


class OpenGIS(BaseResource):
    category = _('Geographic information system')
    service_root_url = models.URLField(_('Service Root URL'), max_length=256)
    query_layer = models.CharField(_('Query Layer'), max_length=256)

    class Meta:
        verbose_name = _('OpenGIS')

    @classmethod
    def get_icon_class(cls):
        return 'gis'

    @endpoint(perm='can_access',
              description=_('Get feature info'),
              parameters={
                  'lat': {'description': _('Latitude'), 'example_value': '45.79689'},
                  'lon': {'description': _('Longitude'), 'example_value': '4.78414'},
              })
    def feature_info(self, request, lat, lon):
        bbox = '%s,%s,%s,%s' % (lat, lon, float(lat) + 0.002, float(lon) + 0.002)
        params = {
            'VERSION': '1.3.0',
            'SERVICE': 'WMS',
            'REQUEST': 'GetFeatureInfo',
            'INFO_FORMAT': 'application/vnd.ogc.gml',
            'STYLES': '',
            'I': '0',
            'J': '0',
            'HEIGHT': '500',
            'WIDTH': '500',
            'CRS': 'EPSG:4171',
            'LAYERS': self.query_layer,
            'QUERY_LAYERS': self.query_layer,
            'BBOX': bbox,
        }
        response = self.requests.get(self.service_root_url, params=params)
        element = ET.fromstring(response.content)
        return {'err': 0, 'data': build_dict_from_xml(element)}
