# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from django.conf import settings
from django.db import models, transaction
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from passerelle.base.models import BaseResource
from passerelle.utils.api import endpoint



def get_tcl_data_url(key):
    if key == 'tclarret':
        return settings.TCL_GEOJSON_URL_TEMPLATE % key
    return settings.TCL_URL_TEMPLATE % key


class Tcl(BaseResource):
    category = _('Transport')

    class Meta:
        verbose_name = _('TCL')

    @endpoint(pattern='^(?P<identifier>\w+)/?$', perm='can_access',
              description=_('Info about a stop'),
              example_pattern='{identifier}/',
              parameters={
                  'identifier': {'description': _('Stop Identifier'), 'example_value': '30211'}
              })
    def stop(self, request, identifier):
        stop_object = Stop.objects.get(id_data=identifier)
        stop = {
            'nom': stop_object.nom,
            'lat': str(stop_object.latitude),
            'lng': str(stop_object.longitude),
            'passings': [],
            'passings_by_line': []
        }
        response = self.requests.get(get_tcl_data_url('tclpassagearret'),
                params={'field': 'id', 'value': identifier})
        response.raise_for_status()

        passings_by_line = {}
        for passing in response.json()['values']:
            try:
                line = Line.objects.filter(code_titan_short=passing['ligne'])[0]
            except IndexError:
                passing['line_info'] = {'ligne': passing['ligne']}
            else:
                passing['line_info'] = {
                    'ligne': line.ligne,
                    'infos': line.infos,
                    'transport_key': line.transport_key,
                    'html_bg_color': line.html_bg_color,
                    'html_fg_color': line.html_fg_color
                }
            stop['passings'].append(passing)
            if not passing['line_info']['ligne'] in passings_by_line:
                passings_by_line[passing['line_info']['ligne']] = []
            passings_by_line[passing['line_info']['ligne']].append(passing)

        stop['passings'].sort(key=lambda x: x['heurepassage'])
        stop['passings_by_line'] = sorted(
                [{'ligne': k, 'line_info': v[0]['line_info'], 'passings': v} for k, v in passings_by_line.items()],
                key=lambda x: x['passings'][0]['heurepassage'])
        return {'data': stop}

    def daily(self):
        super(Tcl, self).daily()
        start_update = timezone.now()

        with transaction.atomic():
            for key in ('tcllignebus', 'tcllignemf', 'tcllignetram'):
                url = get_tcl_data_url(key)
                response = self.requests.get(url)
                response.raise_for_status()
                for line_data in response.json()['values']:
                    line, created = Line.objects.get_or_create(code_titan=line_data['code_titan'])
                    line.__dict__.update(line_data)
                    line.transport_key = key
                    line.save()

            url = get_tcl_data_url('tclarret')
            response = self.requests.get(url)
            response.raise_for_status()
            for feature in response.json()['features']:
                arret_data = feature['properties']
                arret_data['id_data'] = arret_data.pop('id')
                stop, created = Stop.objects.get_or_create(id_data=arret_data['id_data'])
                stop.__dict__.update(arret_data)
                stop.pmr = bool(stop.pmr == 't')
                stop.escalator = bool(stop.escalator == 't')
                stop.ascenseur = bool(stop.ascenseur == 't')
                stop.longitude = feature['geometry']['coordinates'][0]
                stop.latitude = feature['geometry']['coordinates'][1]
                stop.save()

            Line.objects.filter(last_update__lt=start_update).delete()
            Stop.objects.filter(last_update__lt=start_update).delete()


class Line(models.Model):
    transport_key = models.CharField(max_length=20)
    indice = models.CharField(max_length=20, blank=True)
    couleur = models.CharField(max_length=20)
    ut = models.CharField(max_length=10)
    ligne = models.CharField(max_length=10)
    code_titan = models.CharField(max_length=10)
    sens = models.CharField(max_length=10)
    libelle = models.CharField(max_length=100)
    infos = models.CharField(max_length=100, blank=True)

    code_titan_short = models.CharField(max_length=10)
    html_bg_color = models.CharField(max_length=10)
    html_fg_color = models.CharField(max_length=10)

    last_update = models.DateTimeField(_('Last update'), auto_now=True)

    def get_foreground_colour(self, background_colour):
        """Calculates the luminance of the given colour (six hexadecimal digits)
           and returns an appropriate foreground colour."""
        # luminance coefficients taken from section C-9 from
        # http://www.faqs.org/faqs/graphics/colorspace-faq/
        brightess = int(background_colour[0:2], 16) * 0.212671 + \
                int(background_colour[2:4], 16) * 0.715160 + \
                int(background_colour[4:6], 16) * 0.072169
        if brightess > 128:
            fg_colour = '000000'
        else:
            fg_colour = 'ffffff'
        return fg_colour

    def save(self, *args, **kwargs):
        if not self.couleur:
            self.couleur = '255 255 255'
        self.html_bg_color = '%02x%02x%02x' % tuple(int(x) for x in self.couleur.split())
        self.html_fg_color = self.get_foreground_colour(self.html_bg_color)
        self.code_titan_short = re.sub(r'[ab].*', '', self.code_titan)
        return super(Line, self).save(*args, **kwargs)


class Stop(models.Model):
    id_data = models.CharField(max_length=10)
    nom = models.CharField(max_length=50)
    desserte = models.CharField(max_length=200)
    pmr = models.BooleanField(default=False)
    escalator = models.BooleanField(default=False)
    ascenseur = models.BooleanField(default=False)
    latitude = models.FloatField(null=True)
    longitude = models.FloatField(null=True)

    last_update = models.DateTimeField(_('Last update'), auto_now=True)
